# -*- coding: utf-8 -*-
#
# Copyright (c) 2003-2009 by H. Turgut Uyar <uyar@itu.edu.tr>
#
# This file is part of CMFQuestionnaire.
#
# CMFQuestionnaire is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# CMFQuestionnaire is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = """H. Turgut Uyar <uyar@itu.edu.tr>"""
__docformat__ = "restructuredtext"


from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import A4
from reportlab.graphics.shapes import Drawing
from reportlab.graphics.charts.piecharts import Pie
from reportlab.graphics.charts.barcharts import VerticalBarChart

from Products.CMFQuestionnaire import QuestionnaireMessageFactory as _


def putPDFText(pdfReport, text, curX, curY, limitX=550, limitY=30,
               font="Serif", fontSize=10):
    """Output text to the PDF report.

    :Parameters:

      - `pdfReport`: PDF file to write to

      - `text`: text to output

      - `curX`: current X position on the page

      - `curY`: current Y position on the page

      - `limitX`: limit for the X position, default 550

      - `limitY`: limit for the Y position, default 30

      - `font`: font to use for the text, default serif

      - `fontSize`: font size to use for the text, default 10
    """
    tmpText = canvas.textobject.PDFTextObject(pdfReport, x=curX, y=curY)
    pdfText = canvas.textobject.PDFTextObject(pdfReport, x=curX, y=curY)
    tmpText.setFont(font, fontSize)
    pdfText.setFont(font, fontSize)
    for word in text.split(' '):
        if word.strip():
            tmpText.textOut(word + ' ')
            if tmpText.getX() > limitX:
                pdfText.textLine('')
                pdfText.textOut(word + ' ')
                tmpText.textLine('')
                tmpText.textOut(word + ' ')
            else:
                pdfText.textOut(word + ' ')
    pdfText.textLine('')
    curX, curY = pdfText.getX(), pdfText.getY()
    pdfReport.drawText(pdfText)
    return curX, curY


def createPDFReport(title, results, comments, submitCount, pdfFileName,
                    pagesize=A4, leftX=50, topY=770, limitX=550, limitY=30,
                    chartX=270, chartY=140):
    pdfReport = canvas.Canvas(pdfFileName, pagesize=pagesize)
    pdfReport.setTitle(title)
    pdfReport.setAuthor("CMFQuestionnaire")

    curX, curY = putPDFText(pdfReport, title, leftX, topY, fontSize=18)

    curX, curY = putPDFText(pdfReport,
                            "%s: %d" % (_(u"Total submissions"), submitCount),
                            curX, curY, font="Serif-Italic")

    valueStep = 1
    if submitCount > 10:
        valueStep = 5
    if submitCount > 50:
        valueStep = 10
    if submitCount > 100:
        valueStep = 50

    for result in results:
        lines = 1 + (len(title)/100)
        height = lines * 12 + chartY + 20
        if curY - height < limitY:
            pdfReport.showPage()
            curX, curY = leftX, topY

        pdfReport.line(leftX, curY, limitX, curY)
        curX, curY = putPDFText(pdfReport, result['text'], leftX, curY - 20)

        barChart = VerticalBarChart()
        barChart.valueAxis.valueMin = 0
        barChart.valueAxis.valueMax = submitCount + 1
        barChart.valueAxis.valueStep = valueStep

        barChart.categoryAxis.categoryNames = []
        barChart.data = []
        values = []
        i = 0
        for answer in result["answers"]:
            i += 1
            barChart.categoryAxis.categoryNames.append(str(i))
            values.append(answer[0])
        barChart.data.append(values)

        barChart.x = 20
        barChart.y = 20
        barChart.width = 100
        barChart.height = 100

        pieChart = Pie()
        pieChart.labels = barChart.categoryAxis.categoryNames
        pieChart.data = []
        for answer in result["answers"]:
            pieChart.data.append(answer[1])

        pieChart.x = barChart.x + barChart.width + 30
        pieChart.y = barChart.y - 10
        pieChart.width = 100
        pieChart.height = 100

        chart = Drawing(chartX, chartY)
        chart.add(barChart)
        chart.add(pieChart)

        nextY = curY - chartY + 10
        chart.drawOn(pdfReport, leftX, nextY)

        curX, curY = putPDFText(pdfReport,
            _(u"Answer count") + ": " + str(result["answered"]),
            leftX + chartX + 10, curY - 10,
            font="Serif-Italic", fontSize=10)

        curX, curY = putPDFText(pdfReport,
            _(u"Average") + ": " + str(result["average"]),
            leftX + chartX + 10, curY - 10,
            font="Serif-Italic", fontSize=10)
        curY = curY - 10

        legend = result["options"]
        for option in legend:
            curX, curY = putPDFText(pdfReport,
                "%s: %s" % (option['value'], option['text']),
                leftX + chartX + 10, curY,
                font="Serif-Italic", fontSize=10)

        curX, curY = leftX, nextY - 20

    pdfReport.showPage()

    curX, curY = putPDFText(pdfReport, _(u"Comments"), leftX, topY, fontSize=18)
    for comment in comments:
        commentText = comment["text"]
        # estimate a height and check whether we should start a new page
        lines = 1 + (len(commentText) / 100)
        height = lines * 12 + 20
        if curY - height < limitY:
            pdfReport.showPage()
            curX, curY = leftX, topY
        pdfReport.line(leftX, curY, limitX, curY)
        curX, curY = putPDFText(pdfReport, commentText, leftX, curY - 20)

    pdfReport.showPage()
    pdfReport.save()
