# -*- coding: utf-8 -*-
#
# Copyright (c) 2003-2009 by H. Turgut Uyar <uyar@itu.edu.tr>
#
# This file is part of CMFQuestionnaire.
#
# CMFQuestionnaire is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# CMFQuestionnaire is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = """H. Turgut Uyar <uyar@itu.edu.tr>"""
__docformat__ = "restructuredtext"


import Globals

from AccessControl import ClassSecurityInfo
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import _checkPermission
from Products.CMFCore.utils import getToolByName

from Products.Archetypes.atapi import Schema
from Products.Archetypes.atapi import TextField
from Products.Archetypes.atapi import TextAreaWidget
from Products.Archetypes.atapi import BooleanField
from Products.Archetypes.atapi import BooleanWidget
from Products.Archetypes.atapi import LinesField
from Products.Archetypes.atapi import LinesWidget

from Products.ATContentTypes.content.base import ATCTContent
from Products.ATContentTypes.content.base import registerATCT
from Products.ATContentTypes.content.schemata import ATContentTypeSchema
from Products.ATContentTypes.content.schemata import finalizeATCTSchema

from Products.CMFQuestionnaire.config import PROJECTNAME
from Products.CMFQuestionnaire.permissions import ViewQuestionnaire
from Products.CMFQuestionnaire.permissions import ViewQuestionnaireResults
from Products.CMFQuestionnaire.permissions import FilloutQuestionnaire

from Products.CMFQuestionnaire import QuestionnaireMessageFactory as _

from Products.CMFQuestionnaire.config import VISIBLE_BODY_XML
from Products.CMFQuestionnaire.config import VISIBLE_ANSWERS_XML
from Products.CMFQuestionnaire.config import RANDOM_TICKET_SIZE

from elementtree import ElementTree
import operator
import random
from DateTime import DateTime

from Products.CMFQuestionnaire.config import PDF_AVAILABLE
if PDF_AVAILABLE:
    import os
    from StringIO import StringIO
    from Products.CMFQuestionnaire.content.pdf import createPDFReport
    PDF_BASE_DIR = os.path.join(Globals.INSTANCE_HOME, 'var', 'questionnaires')
    try:
        os.makedirs(PDF_BASE_DIR)
    except OSError:
        pass


QuestionnaireSchema = ATContentTypeSchema.copy() + Schema((
    TextField(
        name="body",
        searchable=True,
        allowable_content_types=("txt/xml",),
        default="<questionnaire/>",
        widget=TextAreaWidget(
            label=_(u"label_xml_body", default=u"XML Body"),
            description=_(u"help_xml_body",
                          default=u"Enter the XML body of the questionnaire. You may prefer to use the design tab instead of filling out this field manually."),
            visible=VISIBLE_BODY_XML,
            modes=('edit',),
            ),
        ),

    TextField(
        name="answers",
        allowable_content_types=("txt/xml",),
        default="<answers/>",
        widget=TextAreaWidget(
            label=_(u"label_answers", default=u"Answers"),
            visible=VISIBLE_ANSWERS_XML,
            modes=('edit',),
            ),
        ),

    BooleanField(
        name="use_ticket",
        schemata="tickets",
        default=False,
        widget=BooleanWidget(
            label=_(u"label_use_ticket", default=u"Use Ticket"),
            description=_(u"help_use_ticket",
                          default=u"Do you want to use tickets in addition to roles?"),
            ),
        ),

    LinesField(
        name="available_tickets",
        schemata="tickets",
        widget=LinesWidget(
            label=_(u"label_available_tickets", default=u"Available Tickets"),
            description=_(u"help_available_tickets",
                          default=u"Enter the available tickets. If you did not select to use tickets above, this will be the list of logged-in users who can participate."),
            rows=5,
            ),
        ),
    ),)


finalizeATCTSchema(QuestionnaireSchema)


def _getQuestionInfo(qe, encoding):
    """Return information about a question.
    """
    info = {}
    info['tag'] = qe.tag.encode(encoding)
    info['id'] = qe.get(u'id').encode(encoding)

    te = qe.find(u'text')
    info['text'] = te.text.encode(encoding)

    if hasattr(qe, 'parent'):
        ose = qe.parent.find('legend')
    else:
        ose = qe.find(u'options')

    if not ose:
        info['options'] = []
    else:
        options = [(int(oe.get(u'value')), oe.text) for oe in ose]
        info['options'] = [{'value': str(o[0]), 'text': o[1].encode(encoding)}
                           for o in sorted(options, key=operator.itemgetter(0))]
    return info


def _getGroupInfo(ge, encoding):
    """Return information about a question group.
    """
    info = {}
    info['tag'] = ge.tag.encode(encoding)
    info['id'] = ge.get(u'id').encode(encoding)
    info['scale'] = int(ge.get(u'scale'))

    te = ge.find(u'title')
    info['title'] = te.text.encode(encoding)

    # if the legend node is missing -such as when the group is first
    # created- generate an empty options list
    ose = ge.find(u'legend')
    if ose is None:
        info['options'] = []
    else:
        options = [(int(oe.get(u'value')), oe.text) for oe in ose]
        info['options'] = [{'value': str(o[0]), 'text': o[1].encode(encoding)}
                           for o in sorted(options, key=operator.itemgetter(0))]

    # if the description node is missing -such as when the group is first
    # created- generate an empty description
    de = ge.find(u'description')
    if de is None:
        info['description'] = u''
    else:
        info['description'] = de.text.encode(encoding)
    return info


class Questionnaire(ATCTContent):
    """A questionnaire.
    """

    __implements__ = (ATCTContent.__implements__,)

    _at_rename_after_creation = True

    schema = QuestionnaireSchema

    security = ClassSecurityInfo()

    _encoding = 'UNSET'

    security.declarePrivate("_fromstring")
    def _fromstring(self, field, with_ids=False):
        if Questionnaire._encoding == 'UNSET':
            pt = self.portal_properties
            Questionnaire._encoding = pt.site_properties.getProperty('default_charset')
        try:
            content = str(field)
        except UnicodeEncodeError:
            content = field.encode(self._encoding, 'xmlcharrefreplace')

        if with_ids:
            return ElementTree.XMLID(content)
        else:
            return ElementTree.fromstring(content)

    security.declareProtected(ViewQuestionnaire, "getItems")
    def getItems(self):
        """Return a list of all top-level questions and groups.
        """
        root = self._fromstring(self.body)
        result = []
        for ie in root:
            if ie.tag == 'question':
                result.append(_getQuestionInfo(ie, encoding=self._encoding))
            elif ie.tag == 'group':
                result.append(_getGroupInfo(ie, encoding=self._encoding))
        return result

    security.declareProtected(ModifyPortalContent, "deleteItems")
    def deleteItems(self, ids):
        """Delete the items with the given ids.
        """
        if ids:
            # FIXME: the last item is not deleted
            root = self._fromstring(self.body)
            for e1 in list(root):
                for e2 in e1:
                    if (u'id' in e2.keys()) and (e2.get(u'id') in ids):
                        e1.remove(e2)
                if (u'id' in e1.keys()) and (e1.get(u'id') in ids):
                    root.remove(e1)
            self.body = ElementTree.tostring(root, encoding=self._encoding)

    security.declareProtected(ViewQuestionnaire, "getQuestion")
    def getQuestion(self, id):
        """Return information about the question with the given id.

        :Exceptions:

          - `ValueError`: if a question with the given id can not be found
        """
        root, ids = ElementTree.XMLID(str(self.body))
        qe = ids[id]
        if qe is None:
            raise ValueError, "Question with id %s not found" % id
        return _getQuestionInfo(qe, encoding=self._encoding)

    security.declareProtected(ModifyPortalContent, "addQuestion")
    def addQuestion(self, text, groupId=None):
        """Add a question with the given text to the top-level or to a group.

        Return the id of the newly generated question.

        :Exceptions:

          - `ValueError`: if a group with the given id can not be found
        """
        root, ids = self._fromstring(self.body, with_ids=True)
        if groupId is None:
            parent = root
        else:
            parent = ids[groupId]
            if parent is None:
                raise ValueError, "Group with id %s not found" % groupId

        qids = [int(q[1:]) for q in ids.keys() if ids[q].tag=='question']
        if not qids:
            qid = 1
        else:
            qid = max(qids) + 1
        id = u'q%d' % qid

        qe = ElementTree.SubElement(parent, u'question', id=id)
        te = ElementTree.SubElement(qe, u'text')
        te.text = text.decode(self._encoding)
        self.body = ElementTree.tostring(root, encoding=self._encoding)
        return id.encode(self._encoding)

    security.declareProtected(ModifyPortalContent, "updateQuestion")
    def updateQuestion(self, id, text, options):
        """Update the information about a question.

        :Exceptions:

          - `ValueError`: if a question with the given id can not be found
        """
        root, ids = self._fromstring(self.body, with_ids=True)
        qe = ids[id]
        if qe is None:
            raise ValueError, "Question with id %s not found" % id

        te = qe.find(u'text')
        te.text = text.decode(self._encoding)

        ose = qe.find(u'options')
        if ose is not None:
            qe.remove(ose)
        if options:
            ose = ElementTree.SubElement(qe, u'options')
            options = options.decode(self._encoding)
            options = [o for o in options.splitlines() if o]
            for i in xrange(len(options)):
                oe = ElementTree.SubElement(ose, u'option', value=unicode(i+1))
                oe.text = options[i]
        self.body = ElementTree.tostring(root, encoding=self._encoding)

    security.declareProtected(ViewQuestionnaire, "getGroup")
    def getGroup(self, id):
        """Return information about the group with the given id.

        :Exceptions:

          - `ValueError`: if a group with the given id can not be found
        """
        root, ids = self._fromstring(self.body, with_ids=True)
        ge = ids[id]
        if ge is None:
            raise ValueError, "Group with id %s not found" % id
        return _getGroupInfo(ge, self._encoding)

    security.declareProtected(ModifyPortalContent, "addGroup")
    def addGroup(self, title, scale):
        """Add a group with the given title and scale.

        Return the id of the newly created group.
        """
        root, ids = self._fromstring(self.body, with_ids=True)

        gids = [int(g[1:]) for g in ids.keys() if ids[g].tag=='group']
        if not gids:
            gid = 1
        else:
            gid = max(gids) + 1
        id = u'g%d' % gid

        ge = ElementTree.SubElement(root, u'group', id=id, scale=unicode(scale))
        te = ElementTree.SubElement(ge, u'title')
        te.text = title.decode(self._encoding)
        self.body = ElementTree.tostring(root, encoding=self._encoding)
        return id.encode(self._encoding)

    security.declareProtected(ModifyPortalContent, "updateGroup")
    def updateGroup(self, id, title, scale, legend, description):
        """Update the information about a group.

        :Exceptions:

          - `ValueError`: if a group with the given id can not be found
        """
        root, ids = self._fromstring(self.body, with_ids=True)
        ge = ids[id]
        if ge is None:
            raise ValueError, "Group with id %s not found" % id

        ge.set(u'scale', unicode(scale))

        te = ge.find(u'title')
        te.text = title.decode(self._encoding)

        ose = ge.find(u'legend')
        if ose is not None:
            ge.remove(ose)
        if legend:
            ose = ElementTree.SubElement(ge, u'legend')
            legend = legend.decode(self._encoding)
            options = [o for o in legend.splitlines() if o]
            for i in xrange(len(options)):
                oe = ElementTree.SubElement(ose, u'option', value=unicode(i+1))
                oe.text = options[i]

        de = ge.find(u'description')
        if de is not None:
            ge.remove(de)
        if description:
            de = ElementTree.SubElement(ge, u'description')
            de.text = description.decode(self._encoding)
        self.body = ElementTree.tostring(root, encoding=self._encoding)

    security.declareProtected(ViewQuestionnaire, "getQuestions")
    def getQuestions(self, groupId=None, root=None, ids=None):
        """Return all questions in the questionnaire or in a group.

        :Exceptions:

          - `ValueError`: if a group with the given id can not be found
        """
        if root is None:
            root, ids = self._fromstring(self.body, with_ids=True)
        result = []
        if groupId is None:
            qes = root.getiterator(u'question')
        else:
            ge = ids[groupId]
            if ge is None:
                raise ValueError, "Group with id %s not found" % groupId
            qes = ge.findall(u'question')
        for qe in qes:
            result.append(_getQuestionInfo(qe, encoding=self._encoding))
        return result

    security.declareProtected(ViewQuestionnaire, "isOpen")
    def isOpen(self):
        wt = getToolByName(self, "portal_workflow")
        rs = wt.getInfoFor(self, "review_state")
        if not (rs == "published"):
            return False

        effective = self.EffectiveDate()
        expired = self.ExpirationDate()
        now = DateTime()

        if effective is None:
            isStarted = True
        else:
            isStarted = (now >= DateTime(effective))

        if str(expired) == "None":
            isNotOver = True
        else:
            isNotOver = (now <= DateTime(expired))
        return isStarted and isNotOver

    security.declareProtected(ViewQuestionnaireResults, "getUsedTickets")
    def getUsedTickets(self, sorted=False):
        """Return used tickets.
        """
        aroot, aids = self._fromstring(self.answers, with_ids=True)
        fids = [f.encode(self._encoding) for f in aids if aids[f].tag == 'form']
        if sorted:
            fids.sort()
        return fids

    security.declareProtected(ViewQuestionnaire, "checkTicket")
    def checkTicket(self, ticket):
        """Return whether the given ticket is valid or not.
        """
        # remove persistent lists from existing questionnaires
        if hasattr(self, 'tickets'):
            if hasattr(self, 'used_tickets'):
                tickets = self.used_tickets
            self.available_tickets = tuple(tickets + self.tickets)
            del self.tickets
        if hasattr(self, 'used_tickets'):
            del self.used_tickets

        # TODO: test different ticket use cases
        if not self.use_ticket:
            mt = getToolByName(self, 'portal_membership')
            ticket = mt.getAuthenticatedMember().getId()
        if ticket in self.getUsedTickets():
            return False
        try:
            return ticket in list(self.available_tickets)
        except AttributeError:
            return True

    security.declareProtected(ModifyPortalContent, "createTickets")
    def createTickets(self, count):
        """Create a number of random tickets.

        Return the number of tickets created.
        """
        new_tickets = []
        for i in range(count):
            digits = [str(random.randrange(0, 10))
                      for i in xrange(RANDOM_TICKET_SIZE)]
            ticket = ''.join(digits)
            if (ticket not in self.available_tickets) \
               and (ticket not in new_tickets):
                new_tickets.append(ticket)
        self.available_tickets = self.available_tickets + tuple(new_tickets)
        return count

    security.declareProtected(FilloutQuestionnaire, "fillout")
    def fillout(self, ticket, request):
        """Save a submitted questionnaire.
        """
        if self.checkTicket(ticket):
            root, ids = self._fromstring(self.body, with_ids=True)
            aroot, aids = self._fromstring(self.answers, with_ids=True)
            fe = ElementTree.SubElement(aroot, u'form', id=unicode(ticket))
            qids = [q for q in ids.keys() if ids[q].tag=='question']
            for qid in qids:
                if request.has_key(qid):
                    ae = ElementTree.SubElement(fe, u'question', id=qid)
                    ae.text = request[qid].decode(self._encoding)

            if request.has_key('comment'):
                comment = request['comment'].decode(self._encoding)
                if comment:
                    ce = ElementTree.SubElement(fe, u'comment')
                    ce.text = comment
            self.answers = ElementTree.tostring(aroot, encoding=self._encoding)

    security.declareProtected(ModifyPortalContent, "deleteForms")
    def deleteForms(self, ids):
        """Delete the forms with the given ids.
        """
        if ids:
            aroot, aids = self._fromstring(self.answers, with_ids=True)
            for id in ids:
                fe = aids[id]
                if fe is None:
                    raise ValueError, "Form with id %s can not be found" % id
                aroot.remove(fe)
            self.answers = ElementTree.tostring(aroot, encoding=self._encoding)

    security.declareProtected(ViewQuestionnaire, "getAnswer")
    def getAnswer(self, ticket, questionId, aroot=None, aids=None):
        """Return the answer for a question in a form.
        """
        if not _checkPermission(ViewQuestionnaireResults, self):
            return None
        if (not ticket) or (not ticket in self.getUsedTickets()):
            return None

        if aroot is None:
            aroot, aids = self._fromstring(self.answers, with_ids=True)

        # XXX: form ids might interfere with question/group ids
        fe = aids[unicode(ticket)]
        if fe is not None:
            for qe in fe.findall(u'question'):
                if qe.get(u'id') == questionId:
                    return qe.text.encode(self._encoding)
        return None

    security.declareProtected(ViewQuestionnaire, "getComment")
    def getComment(self, ticket):
        """Return the comment in a form.
        """
        if not _checkPermission(ViewQuestionnaireResults, self):
            return None
        if (not ticket) or (not ticket in self.getUsedTickets()):
            return None

        aroot, aids = self._fromstring(self.answers, with_ids=True)
        # XXX: form ids might interfere with question/group ids
        fe = aids[unicode(ticket)]
        if fe is not None:
            ce = fe.find(u'comment')
            if ce is not None:
                return ce.text.encode(self._encoding)
        return None

    security.declarePrivate("_getAnswersByQuestionId")
    def _getAnswersByQuestionId(self, questionId, aroot=None):
        """Return the answers in all submitted forms for a question.
        """
        if aroot is None:
            aroot = self._fromstring(self.answers)
        answers = []
        for qe in aroot.getiterator(u'question'):
            if qe.get(u'id') == questionId:
                answer = qe.text.encode(self._encoding)
                answers.append(answer)
        return answers

    security.declareProtected(ViewQuestionnaireResults, "evaluate")
    def evaluate(self):
        """Return the answer counts and percentages for each question.
        """
        # remove persistent summary dict from existing questionnaires
        if hasattr(self, 'summary'):
            del self.summary

        root, ids = self._fromstring(self.body, with_ids=True)
        for ge in root.findall(u'group'):
            for qe in ge.findall(u'question'):
                qe.parent = ge

        aroot = self._fromstring(self.answers)

        result = []
        for question in self.getQuestions(root=root, ids=ids):
            answered = 0
            counts = {}
            for answer in self._getAnswersByQuestionId(question.get('id'), aroot):
                if answer:
                    answered += 1
                    if not counts.has_key(answer):
                        counts[answer] = 1
                    else:
                        counts[answer] += 1

            for option in question.get('options'):
                v = option.get('value')
                if not counts.has_key(v):
                    counts[v] = 0

            # compute the percentages of each answer and the weighted average
            average = 0
            total = 0
            summary = []
            for answer in counts.keys():
                percent = 100 * counts[answer] / answered
                summary.append((int(answer), counts[answer], percent))
                total += int(answer) * counts[answer]
            if total > 0:
                average = round(float(total) / answered, 2)

            question['answered'] = answered
            question['average'] = average
            question['answers'] = sorted(summary, key=operator.itemgetter(0))
            result.append(question)
        return result

    security.declareProtected(ViewQuestionnaireResults, "getComments")
    def getComments(self):
        """Return all comments.
        """
        aroot = self._fromstring(self.answers)
        result = []
        for fe in aroot.findall(u'form'):
            ce = fe.find(u'comment')
            if ce is not None:
                comment = {}
                comment['id'] = fe.get(u'id')
                comment['text'] = ce.text.encode(self._encoding)
                result.append(comment)
        return result

    security.declareProtected(ViewQuestionnaireResults, "isPDFAvailable")
    def isPDFAvailable(self):
        """Return whether PDF report generation is available.
        """
        return PDF_AVAILABLE

    security.declareProtected(ViewQuestionnaireResults, "getXMLDump")
    def getXMLDump(self):
        """Return an XML text containing all questionnaire data.
        """
        root = self._fromstring(self.body)
        aroot = self._fromstring(self.answers)
        root.append(aroot)
        return ElementTree.tostring(root, encoding=self._encoding)

    security.declareProtected(ViewQuestionnaireResults, "getCSVDump")
    def getCSVDump(self):
        """Return a CSV text containing all questionnaire data.
        """
        root, ids = self._fromstring(self.body, with_ids=True)
        aroot, aids = self._fromstring(self.answers, with_ids=True)
        qids = sorted([int(q[1:]) for q in ids.keys() if ids[q].tag=='question'])
        qids = [u'q%d' % q for q in qids]
        csvString = u"Id,%s\n" % u','.join(qids)
        for ticket in self.getUsedTickets():
            csvString += unicode(ticket)
            for qid in qids:
                answer = self.getAnswer(ticket, qid, aroot=aroot, aids=aids)
                if answer:
                    csvString += u",%s" % unicode(answer)
                else:
                    csvString += u","
            csvString += u"\n"
        return csvString.encode(self._encoding)

    security.declareProtected(ViewQuestionnaireResults, "getPDFReport")
    def getPDFReport(self):
        """Return a PDF report of the results.
        """
        if not self.isPDFAvailable():
            raise ImportError, "Reportlab PDF modules not installed"

        results = self.evaluate()
        comments = self.getComments()
        submitCount = len(self.getUsedTickets())
        pdfFileName = os.path.join(PDF_BASE_DIR, "%s.pdf" % self.id)
        pdfFile = createPDFReport(self.title, results, comments,
                                  submitCount, pdfFileName)
        pdfFile = open(pdfFileName)
        pdfString = StringIO(pdfFile.read())
        pdfFile.close()
        return pdfString.read()


registerATCT(Questionnaire, PROJECTNAME)
